/*
 * boidflocks.cpp
 *
 *    (C) 2004-2005, End of the World Production, LLC.,
 *     & Alex S, a/k/a Particle, a/k/a Prof.Phreak
 *
 *       alex@theparticle.com
 *       http://www.theparticle.com
 */

/*
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

 See the GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/




/**********************************************************************
 * without regard for any sense of design or code esthetics, we'll
 * just dump everything into one huge and unmaintanable file.
 **********************************************************************/


#include <windows.h>
#include <scrnsave.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>

#include "resource.h"


#pragma comment(lib, "scrnsave.lib")
#pragma comment(lib, "user32.lib")
#pragma comment(lib, "advapi32.lib")
#pragma comment(lib, "gdi32.lib")
#pragma comment(lib, "comctl32.lib")


#define PI 3.14159265358979323846264338327950288419716939937510

/**********************************************************************
 * random generator
 **********************************************************************/
inline double random(){
    return (rand() / (double)RAND_MAX);
}

/**********************************************************************
 * The moving object (boid!)
 **********************************************************************/
class MovingObject {
  public:   // screw incapsulation

    double x,y;
    double direction;
    double speed;

    inline MovingObject(){
        x = y = direction = speed = 0;        
    }

    inline MovingObject(double x_,double y_,double direction_,double speed_){
        x = x_;
        y = y_;
        direction = direction_;
        speed = speed_;
    }

    void move(int w,int h){
        double Sin = sin(direction);
        double Cos = cos(direction);
        double nx = x + Cos * speed;
        double ny = y + Sin * speed;
        if(nx > w-1){
            nx = w-1;
            Cos = -Cos;
            speed += 0.5*(random()-0.5);
        }
        if(nx < 0){
            nx = 0;
            Cos = -Cos;
            speed += 0.5*(random()-0.5);
        }
        if(ny > h-1){
            ny = h-1;
            Sin = -Sin;
            speed += 0.5*(random()-0.5);
        }
        if(ny < 0){
            ny = 0;
            Sin = -Sin;
            speed += 0.5*(random()-0.5);
        }
        if(speed < 0.1)
            speed = 0.1;
        if(speed > 10)
            speed = 10;
        direction = atan2(Sin,Cos);
        x = nx;
        y = ny;
    }
};


/**********************************************************************
 * bitmap stuff
 **********************************************************************/

struct pBITMAPINFO {
    BITMAPINFOHEADER bmiHeader;
    RGBQUAD bmiColors[256];
}BMInfo;

struct pLOGPALETTE {
    WORD    palVersion;
    WORD    palNumEntries;
    PALETTEENTRY palPalEntry[256];
}PalInfo;

/**********************************************************************
 * GLOBALS
 **********************************************************************/

HBITMAP hBM;
BYTE* double_buffer;
BYTE* double_double_buffer;

// buffer size
#define DB_WIDTH            320
#define DB_HEIGHT           240

#define NUM_OBJECTS         200
#define COLLISION_RADIUS    2.5
#define GRID_WIDTH          ((int)(DB_WIDTH / (COLLISION_RADIUS * 2)))
#define GRID_HEIGHT         ((int)(DB_HEIGHT / (COLLISION_RADIUS * 2)))
#define EPSILON             0.001

int queue[GRID_WIDTH * GRID_HEIGHT];
int queuelast[GRID_WIDTH * GRID_HEIGHT];
int next[NUM_OBJECTS];
int objx[NUM_OBJECTS];
int objy[NUM_OBJECTS];

MovingObject objects[NUM_OBJECTS];

/**********************************************************************
 * initialize objects
 **********************************************************************/
void objects_init(){
    for(int i=0;i<NUM_OBJECTS;i++){
        objects[i].x = random()*DB_WIDTH;
        objects[i].y = random()*DB_HEIGHT;
        objects[i].direction = random()*(2.0 * PI);
        objects[i].speed = random()*COLLISION_RADIUS/2.0;
    }
}

/**********************************************************************
 * given current object positions and speed, move'em one time step
 **********************************************************************/
void objects_update(){
    int i,j;
    for(i=0;i<GRID_WIDTH*GRID_HEIGHT;i++){
        queue[i] = -1;
    }
    double avgx = 0;
    double avgy = 0;
    for(i=0;i<NUM_OBJECTS;i++){
        objects[i].move(DB_WIDTH-1,DB_HEIGHT-1);
        
        avgx += objects[i].x;
        avgy += objects[i].y;

        int x = objx[i] = (int)( GRID_WIDTH*(objects[i].x / DB_WIDTH) );
        int y = objy[i] = (int)( GRID_HEIGHT*(objects[i].y / DB_HEIGHT) );            
        int bucket = y * GRID_WIDTH + x;

        // is there anything in this bucket already
        if(queue[bucket]==-1){  //  no
            //the start/end of this bucket is set to index of element.
            queuelast[bucket] = queue[bucket] = i;
        }else{                  // yes
            // the last pointer in this bucket will be this index
            queuelast[bucket] = next[ queuelast[bucket] ] = i;
        }
        // initialize the next pointer for that element.
        next[i]=-1;
        
    }
    avgx /= NUM_OBJECTS;
    avgy /= NUM_OBJECTS;

    double diam = (COLLISION_RADIUS + COLLISION_RADIUS);
    double distsquared = (COLLISION_RADIUS + COLLISION_RADIUS)*
                        (COLLISION_RADIUS + COLLISION_RADIUS);   //diam * diam;

    for(i=0;i<NUM_OBJECTS;i++){
        double x = objects[i].x;
        double y = objects[i].y;
        int bx = objx[i];
        int by = objy[i];
            
        int Ux = bx - 1 < 0 ? 0:bx-1;
        int Uy = by - 1 < 0 ? 0:by-1;
        int Lx = bx + 1 > GRID_WIDTH-1 ? GRID_WIDTH-1:bx+1;
        int Ly = by + 1 > GRID_HEIGHT-1 ? GRID_HEIGHT-1:by+1;

        // direction stats.
        double _avgsin = 0;
        double _avgcos = 0;
        double _avgx = 0;
        double _avgy = 0;
        double _avgspeed = 0;
        int _numobjects = 0;

        for(int cy=Uy;cy<=Ly;cy++){
            for(int cx=Ux;cx<=Lx;cx++){
                int bucket = cy * GRID_WIDTH + cx;
                for(j = queue[bucket];j != -1; j = next[j]){
                    if(i != j){
                        double ox = objects[j].x;
                        double oy = objects[j].y;
                        double dx = ox - x;
                        double dy = oy - y;
                        double dd = dx * dx + dy * dy;
                            
                        if(dd < distsquared){
                            double dist = sqrt(dd) - EPSILON;
                            double disp = ((diam - dist)/2)/dist;

                            objects[i].x -= dx * disp;
                            objects[i].y -= dy * disp;

                            objects[j].x += dx * disp;
                            objects[j].y += dy * disp;
                        }                            
                    }

                    _avgsin += sin(objects[j].direction);
                    _avgcos += cos(objects[j].direction);
                    _avgspeed += objects[j].speed;
                    _avgx += objects[j].x;
                    _avgy += objects[j].y;
                    _numobjects++;
                }
            }
        }

        // adjust current object according to averages.
        _avgsin /= _numobjects;
        _avgcos /= _numobjects;
        _avgx /= _numobjects;
        _avgy /= _numobjects;
        _avgspeed /= _numobjects;

        // add some noise into the motion
        _avgsin += 0.2*(random()-0.5);
        _avgcos += 0.2*(random()-0.5);
        objects[i].speed += 0.2*(random()-0.5);

        // find heading into center of local group.
        double _dx = _avgx - objects[i].x ;
        double _dy = _avgy - objects[i].y ;
        // normalize.
        double _ddist = sqrt(_dx * _dx + _dy * _dy);

        // find heading into center of -global- group.
        double _dxall = avgx - objects[i].x;
        double _dyall = avgy - objects[i].y;
        double _ddistall = sqrt(_dxall * _dxall + _dyall * _dyall);
        if(_ddistall == 0)
            _ddistall = EPSILON;
        _dxall /= _ddistall;
        _dyall /= _ddistall;

        // if it's getting a bit too crowded, get away from center.
        if(_numobjects > 8){   //adjust number for different flocks.             
            if(_ddist > 0){
                _dx /= _ddist;
                _dy /= _ddist;
                objects[i].direction = atan2(_avgsin*0.3 + -_dy*0.7,_avgcos*0.3 + -_dx*0.7);
            }else{
                objects[i].direction = atan2(_avgsin,_avgcos);
            }
        }else{
            if(_ddist > 0){
                _dx /= _ddist;
                _dy /= _ddist;
                objects[i].direction = atan2(
                    _avgsin*0.8 + _dy*0.195 + _dyall*0.005,
                    _avgcos*0.8 + _dx*0.195 + _dxall*0.005
                );
            }else{
                objects[i].direction = atan2(_avgsin,_avgcos);
            }
        }
        objects[i].speed += 0.01*(_avgspeed - objects[i].speed);
    }
}





/**********************************************************************
 * plot a big `dot'
 **********************************************************************/
inline void plot(int dotx,int doty){
    // plot our `moving' dot.

    if(dotx > DB_WIDTH-1)
        dotx = DB_WIDTH;
    if(dotx < 1)
        dotx = 1;
    if(doty > DB_HEIGHT-1)
        doty = DB_HEIGHT;
    if(doty < 1)
        doty = 1;

    double_double_buffer[(doty-1)*DB_WIDTH+(dotx-1)] = 255;
    double_double_buffer[(doty-1)*DB_WIDTH+(dotx)] = 255;
    double_double_buffer[(doty-1)*DB_WIDTH+(dotx+1)] = 255;
    double_double_buffer[doty*DB_WIDTH+(dotx-1)] = 255;
    double_double_buffer[doty*DB_WIDTH+dotx] = 255;
    double_double_buffer[doty*DB_WIDTH+(dotx+1)] = 255;
    double_double_buffer[(doty+1)*DB_WIDTH+(dotx-1)] = 255;
    double_double_buffer[(doty+1)*DB_WIDTH+dotx] = 255;
    double_double_buffer[(doty+1)*DB_WIDTH+(dotx+1)] = 255;
}



/**********************************************************************
 * function that gets called every few milliseconds
 **********************************************************************/
void p_update(){

    int x,y,i;
    
    objects_update();

    for(i=0;i<NUM_OBJECTS;i++){
        plot(objects[i].x,objects[i].y);
    }



    // do a `blur'
    for (y=1;y<DB_HEIGHT-1;y++) {
        for (x=1;x<DB_WIDTH-1;x++) {
            int sum =
                +double_double_buffer[(y+-1)*DB_WIDTH+(x+-1)]
                +double_double_buffer[(y+-1)*DB_WIDTH+(x)]
                +double_double_buffer[(y+-1)*DB_WIDTH+(x+1)]
                +double_double_buffer[(y)*DB_WIDTH+(x+-1)]
                +double_double_buffer[(y)*DB_WIDTH+(x)] * 2
                +double_double_buffer[(y)*DB_WIDTH+(x+1)]
                +double_double_buffer[(y+1)*DB_WIDTH+(x+-1)]
                +double_double_buffer[(y+1)*DB_WIDTH+(x)]
                +double_double_buffer[(y+1)*DB_WIDTH+(x+1)];
            sum /= 10.4;
            double_buffer[y*DB_WIDTH+x] = (BYTE)sum;
        }
    }
    memcpy(double_double_buffer,double_buffer,sizeof(BYTE)*DB_WIDTH*DB_HEIGHT);
}

/**********************************************************************
 * function that figures out the palette (mostly yellowish).
 **********************************************************************/
void getpalette(RGBQUAD* p){
    int i;

    p[0].rgbRed = (int)(0);
    p[0].rgbGreen = (int)(0);
    p[0].rgbBlue = (int)(0);
    p[0].rgbReserved = 0;

    for (i=1;i<256;i++) {
        p[i].rgbRed = i;
        p[i].rgbGreen = i;
        p[i].rgbBlue = i/4;
        p[i].rgbReserved = 0;
    }
}

/**********************************************************************
 * initialize double buffer
 **********************************************************************/
void init_double_buffer(HDC hDC){
    HPALETTE PalHan;
    RGBQUAD palette[256];
    int i;

    double_double_buffer = (BYTE*)malloc(sizeof(BYTE)*DB_WIDTH*DB_HEIGHT);
    memset(double_double_buffer,0,DB_WIDTH*DB_HEIGHT);

    BMInfo.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    BMInfo.bmiHeader.biWidth = DB_WIDTH;
    BMInfo.bmiHeader.biHeight = -abs(DB_HEIGHT);
    BMInfo.bmiHeader.biPlanes = 1;
    BMInfo.bmiHeader.biBitCount = 8;
    BMInfo.bmiHeader.biCompression = BI_RGB;
    BMInfo.bmiHeader.biSizeImage = 0;
    BMInfo.bmiHeader.biXPelsPerMeter = 0;
    BMInfo.bmiHeader.biYPelsPerMeter = 0;
    BMInfo.bmiHeader.biClrUsed = 256;
    BMInfo.bmiHeader.biClrImportant = 256;

    getpalette(palette);

    for (i=0;i<256;i++)
        BMInfo.bmiColors[i] = palette[i];

    PalInfo.palVersion = 0x300;
    PalInfo.palNumEntries = 256;
    for (i=0;i<256;i++) {
        PalInfo.palPalEntry[i].peRed = palette[i].rgbRed;
        PalInfo.palPalEntry[i].peGreen = palette[i].rgbGreen;
        PalInfo.palPalEntry[i].peBlue = palette[i].rgbBlue;
        PalInfo.palPalEntry[i].peFlags = PC_NOCOLLAPSE;
    }
    PalHan = CreatePalette((LOGPALETTE*)&PalInfo);
    SelectPalette(hDC,PalHan,FALSE);
    RealizePalette(hDC);
    DeleteObject(PalHan);
    hBM = CreateDIBSection(hDC,(BITMAPINFO*)&BMInfo,
                           DIB_RGB_COLORS,(void**)&double_buffer,0,0);
}

/**********************************************************************
 * copy double buffer to screen; stretching
 **********************************************************************/
void blit(HDC hDC,RECT Dest){
    HDC Context;
    HBITMAP DefaultBitmap;
    Context = CreateCompatibleDC(0);
    DefaultBitmap = (HBITMAP)SelectObject(Context,hBM);
    StretchBlt(hDC,0,0,Dest.right,Dest.bottom,Context,0,0,DB_WIDTH,DB_HEIGHT,SRCCOPY);
    SelectObject(Context,DefaultBitmap);
    DeleteDC(Context);
    DeleteObject(DefaultBitmap);
}

/**********************************************************************
 * proc, initialize screen saver
 **********************************************************************/
LRESULT WINAPI ScreenSaverProc(HWND hwnd,UINT message,WPARAM wParam,LPARAM lParam){
    static HDC hdc,memdc;
    static RECT rect;
    static unsigned int timer;
    static HBITMAP hBitmap;
    static BYTE* double_buffer;

    switch (message) {
    case WM_CREATE:
        srand(time(NULL));
        objects_init();
        hdc = GetDC(hwnd);
        GetClientRect(hwnd,&rect);        
        init_double_buffer(hdc);
        ReleaseDC(hwnd,hdc);
        timer = SetTimer(hwnd,1,33,NULL);
        break;
    case WM_TIMER:
        p_update();
        hdc = GetDC(hwnd);
        GetClientRect(hwnd,&rect);
        blit(hdc,rect);
        ReleaseDC(hwnd,hdc);
        break;
    case WM_ERASEBKGND:
        break;
    case WM_DESTROY:
        KillTimer(hwnd,timer);
        break;
    default:
        return DefScreenSaverProc(hwnd,message,wParam,lParam);
    }
    return 0;
}


BOOL WINAPI ScreenSaverConfigureDialog(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
    return FALSE;
}

BOOL WINAPI RegisterDialogClasses(HANDLE hInst)
{
    return TRUE;
} 

